;; This procedure returns a new list
;; containing the elements in the
;; original for which pred is true
(define (filter pred lst)
  (cond ((null? lst) nil)
        ((pred (car lst))
         (cons (car lst)
               (filter pred (cdr lst))))
        (else (filter pred (cdr lst)))))

;;; Base methods
(define empty-set nil)

(define (set-contains? elm set)
  (cond ((null? set) #f)
    ((equal? (car set) elm) #t)
    (else (set-contains? elm (cdr set)))))

(define (set-add elm set)
  (if (set-contains? elm set)
      set
      (cons elm set)))

;; Tests
(define s (set-add 3 (set-add 2 (set-add 1 empty-set))))
(set-contains? 1 s)
(set-contains? 2 s)
(set-contains? 3 s)
(set-contains? 4 s)


;;; Conversion between sets and lists (constructors)
(define (list-to-set lst)
  (fold-right set-add empty-set lst))
(define (set-to-list set) set)

;; Tests
(define l (list-to-set (list 1 2 3 4)))

;;; Abstract operators
(define (set-remove elm set)
  (list-to-set
   (filter (lambda (x) (not (= elm x)))
       (set-to-list set))))
(set-remove 3 (list-to-set (list 1 2 3 4 5)))

(define sym-type 0)
(define num-type 1)
(define str-type 2)
(define (type-enum v)
  (cond ((symbol? v) sym-type)
	((number? v) num-type)
	((string? v) str-type)
	(else (error "Can't (yet) compare your type")))
  
;; Provides a total ordering on numbers and symbols
(define (super<? a b)
  (let ((sym-type 0) (num-type 1) (str-type 2))

      (let ((a-type (type-enum a))
	    (b-type (type-enum b)))
	(if (not (= a-type b-type))
	    (< a-type b-type)
	    (cond ((= a-type sym-type) (symbol<? a b))
		  ((= a-type num-type) (< a b))
		  ((= a-type str-type) (string<? a b))))))))

; Allowed: basic list ops, filter, fold-right, map, filter, append, length, sort
(define (set-eq? s1 s2)
  (define (helper l1 l2)
    (cond ((null? l1) #t)
      ((equal? (car l1) (car l2))
       (helper (cdr l1) (cdr l2)))
      (else #f)))
  (let ((l1 (sort (set-to-list s1) <))
    (l2 (sort (set-to-list s2) <)))
    (if (= (length l1) (length l2))
    (helper l1 l2)
    #f)))
(define (set-eq? s1 s2)
  (define (bool-and a b) (and a b))
  (let ((l1 (sort (set-to-list s1) <))
    (l2 (sort (set-to-list s2) <)))
    (if (= (length l1) (length l2))
    (fold-right bool-and #t (map equal? l1 l2))
    #f)))
(set-eq? (list-to-set (list 1 2 3)) (list-to-set (list 1 2 3))) ; -> #t
(set-eq? (list-to-set (list 1 2 3)) (list-to-set (list 3 2 1))) ; -> #t
(set-eq? (list-to-set (list 1 2 3)) (list-to-set (list 1 2 3 4))) ; -> #f
(set-eq? (list-to-set (list 1 2 3)) (list-to-set (list 1 2))) ; -> #f
(set-eq? (list-to-set (list )) (list-to-set (list 1 2 3))) ; -> #f
(set-eq? empty-set empty-set) ; -> #f

(define (set-union s1 s2)
  (fold-right set-add s2 (set-to-list s1)))
(set-to-list (set-union empty-set empty-set))
(set-to-list (set-union empty-set (list-to-set (list 1 1 2))))
(set-to-list (set-union (list-to-set (list 1 2 3)) (list-to-set (list 1 1 2))))

(define (set-intersection s1 s2)
  (list-to-set (filter (lambda (elm) (set-contains? elm s2)) (set-to-list s1))))
(define s1 empty-set)
(define s2 (list-to-set (list 1 1 2)))
(define s3 (list-to-set (list 1 2 3)))
(define s4 (list-to-set (list 2 3 4)))
(define s5 (list-to-set (list 7 8 9)))
(set-eq? (set-intersection s1 s1) empty-set)
(set-eq? (set-intersection s1 s2) empty-set)
(set-eq? (set-intersection s2 s2) s2)
(set-eq? (set-intersection s2 s3) (list-to-set (list 1 2)))
(set-eq? (set-intersection s3 s4) (list-to-set (list 2 3)))
(set-eq? (set-intersection s2 s5) empty-set)

; Returns a set that has all elements of set s1 that are not in set s2
(define (set-diff s1 s2)
  (list-to-set
   (filter (lambda (elm) (not (set-contains? elm s2)))
       (set-to-list s1))))
(set-eq? (set-diff s1 s1) empty-set)
(set-eq? (set-diff s2 s1) s2)
(set-eq? (set-diff s3 s2) (list-to-set (list 3)))
(set-eq? (set-diff s4 s2) (list-to-set (list 4 3)))
(set-eq? (set-diff s5 s2) s5)








